/****************************************************************************
** Copyright (C) 2020 MikroElektronika d.o.o.
** Contact: https://www.mikroe.com/contact
**
** Permission is hereby granted, free of charge, to any person obtaining a copy
** of this software and associated documentation files (the "Software"), to deal
** in the Software without restriction, including without limitation the rights
** to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, subject to the following conditions:
** The above copyright notice and this permission notice shall be
** included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
** DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
**  USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

/*!
 * @file c8800retro.c
 * @brief 8800 Retro Click Driver.
 */

#include "c8800retro.h"

/** 
 * @brief 8800 Retro Click IBM BIOS 8x8 font array.
 * @details FONT NAME "IBM BIOS 8x8"
 *          0-31,127 : Code page 437, cf. https://en.wikipedia.org/wiki/Code_page_437
 *          32-126   : ASCII
 *          160-255  : ISO-8859-1, cf. https://en.wikipedia.org/wiki/ISO/IEC_8859-1
*/
const uint8_t font8x8_ib8x8u[ 224 ][ 8 ] = {
    // ---------------------------  0-127 ---------------------------
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0x00 (uni0000.dup1)
    { 0x7E, 0x81, 0xA5, 0x81, 0xBD, 0x99, 0x81, 0x7E },  // 0x01 (uni0001)
    { 0x7E, 0xFF, 0xDB, 0xFF, 0xC3, 0xE7, 0xFF, 0x7E },  // 0x02 (uni0002)
    { 0x6C, 0xFE, 0xFE, 0xFE, 0x7C, 0x38, 0x10, 0x00 },  // 0x03 (uni0003)
    { 0x10, 0x38, 0x7C, 0xFE, 0x7C, 0x38, 0x10, 0x00 },  // 0x04 (uni0004)
    { 0x38, 0x7C, 0x38, 0xFE, 0xFE, 0x7C, 0x38, 0x7C },  // 0x05 (uni0005)
    { 0x10, 0x10, 0x38, 0x7C, 0xFE, 0x7C, 0x38, 0x7C },  // 0x06 (uni0006)
    { 0x00, 0x00, 0x18, 0x3C, 0x3C, 0x18, 0x00, 0x00 },  // 0x07 (uni0007)
    { 0xFF, 0xFF, 0xE7, 0xC3, 0xC3, 0xE7, 0xFF, 0xFF },  // 0x08 (uni0008)
    { 0x00, 0x3C, 0x66, 0x42, 0x42, 0x66, 0x3C, 0x00 },  // 0x09 (uni0009)
    { 0xFF, 0xC3, 0x99, 0xBD, 0xBD, 0x99, 0xC3, 0xFF },  // 0x0A (uni000A)
    { 0x0F, 0x07, 0x0F, 0x7D, 0xCC, 0xCC, 0xCC, 0x78 },  // 0x0B (uni000B)
    { 0x3C, 0x66, 0x66, 0x66, 0x3C, 0x18, 0x7E, 0x18 },  // 0x0C (uni000C)
    { 0x3F, 0x33, 0x3F, 0x30, 0x30, 0x70, 0xF0, 0xE0 },  // 0x0D (uni000D)
    { 0x7F, 0x63, 0x7F, 0x63, 0x63, 0x67, 0xE6, 0xC0 },  // 0x0E (uni000E)
    { 0x99, 0x5A, 0x3C, 0xE7, 0xE7, 0x3C, 0x5A, 0x99 },  // 0x0F (uni000F)
    { 0x80, 0xE0, 0xF8, 0xFE, 0xF8, 0xE0, 0x80, 0x00 },  // 0x10 (uni0010)
    { 0x02, 0x0E, 0x3E, 0xFE, 0x3E, 0x0E, 0x02, 0x00 },  // 0x11 (uni0011)
    { 0x18, 0x3C, 0x7E, 0x18, 0x18, 0x7E, 0x3C, 0x18 },  // 0x12 (uni0012)
    { 0x66, 0x66, 0x66, 0x66, 0x66, 0x00, 0x66, 0x00 },  // 0x13 (uni0013)
    { 0x7F, 0xDB, 0xDB, 0x7B, 0x1B, 0x1B, 0x1B, 0x00 },  // 0x14 (uni0014)
    { 0x3E, 0x63, 0x38, 0x6C, 0x6C, 0x38, 0xCC, 0x78 },  // 0x15 (uni0015)
    { 0x00, 0x00, 0x00, 0x00, 0x7E, 0x7E, 0x7E, 0x00 },  // 0x16 (uni0016)
    { 0x18, 0x3C, 0x7E, 0x18, 0x7E, 0x3C, 0x18, 0xFF },  // 0x17 (uni0017)
    { 0x18, 0x3C, 0x7E, 0x18, 0x18, 0x18, 0x18, 0x00 },  // 0x18 (uni0018)
    { 0x18, 0x18, 0x18, 0x18, 0x7E, 0x3C, 0x18, 0x00 },  // 0x19 (uni0019)
    { 0x00, 0x18, 0x0C, 0xFE, 0x0C, 0x18, 0x00, 0x00 },  // 0x1A (uni001A)
    { 0x00, 0x30, 0x60, 0xFE, 0x60, 0x30, 0x00, 0x00 },  // 0x1B (uni001B)
    { 0x00, 0x00, 0xC0, 0xC0, 0xC0, 0xFE, 0x00, 0x00 },  // 0x1C (uni001C)
    { 0x00, 0x24, 0x66, 0xFF, 0x66, 0x24, 0x00, 0x00 },  // 0x1D (uni001D)
    { 0x00, 0x18, 0x3C, 0x7E, 0xFF, 0xFF, 0x00, 0x00 },  // 0x1E (uni001E)
    { 0x00, 0xFF, 0xFF, 0x7E, 0x3C, 0x18, 0x00, 0x00 },  // 0x1F (uni001F)
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0x20 (space)
    { 0x30, 0x78, 0x78, 0x30, 0x30, 0x00, 0x30, 0x00 },  // 0x21 (exclam)
    { 0x6C, 0x6C, 0x6C, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0x22 (quotedbl)
    { 0x6C, 0x6C, 0xFE, 0x6C, 0xFE, 0x6C, 0x6C, 0x00 },  // 0x23 (numbersign)
    { 0x30, 0x7C, 0xC0, 0x78, 0x0C, 0xF8, 0x30, 0x00 },  // 0x24 (dollar)
    { 0x00, 0xC6, 0xCC, 0x18, 0x30, 0x66, 0xC6, 0x00 },  // 0x25 (percent)
    { 0x38, 0x6C, 0x38, 0x76, 0xDC, 0xCC, 0x76, 0x00 },  // 0x26 (ampersand)
    { 0x60, 0x60, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0x27 (quotesingle)
    { 0x18, 0x30, 0x60, 0x60, 0x60, 0x30, 0x18, 0x00 },  // 0x28 (parenleft)
    { 0x60, 0x30, 0x18, 0x18, 0x18, 0x30, 0x60, 0x00 },  // 0x29 (parenright)
    { 0x00, 0x66, 0x3C, 0xFF, 0x3C, 0x66, 0x00, 0x00 },  // 0x2A (asterisk)
    { 0x00, 0x30, 0x30, 0xFC, 0x30, 0x30, 0x00, 0x00 },  // 0x2B (plus)
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x30, 0x60 },  // 0x2C (comma)
    { 0x00, 0x00, 0x00, 0xFC, 0x00, 0x00, 0x00, 0x00 },  // 0x2D (hyphen)
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x30, 0x00 },  // 0x2E (period)
    { 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0, 0x80, 0x00 },  // 0x2F (slash)
    { 0x7C, 0xC6, 0xCE, 0xDE, 0xF6, 0xE6, 0x7C, 0x00 },  // 0x30 (zero)
    { 0x30, 0x70, 0x30, 0x30, 0x30, 0x30, 0xFC, 0x00 },  // 0x31 (one)
    { 0x78, 0xCC, 0x0C, 0x38, 0x60, 0xC4, 0xFC, 0x00 },  // 0x32 (two)
    { 0x78, 0xCC, 0x0C, 0x38, 0x0C, 0xCC, 0x78, 0x00 },  // 0x33 (three)
    { 0x1C, 0x3C, 0x6C, 0xCC, 0xFE, 0x0C, 0x1E, 0x00 },  // 0x34 (four)
    { 0xFC, 0xC0, 0xF8, 0x0C, 0x0C, 0xCC, 0x78, 0x00 },  // 0x35 (five)
    { 0x38, 0x60, 0xC0, 0xF8, 0xCC, 0xCC, 0x78, 0x00 },  // 0x36 (six)
    { 0xFC, 0xCC, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x00 },  // 0x37 (seven)
    { 0x78, 0xCC, 0xCC, 0x78, 0xCC, 0xCC, 0x78, 0x00 },  // 0x38 (eight)
    { 0x78, 0xCC, 0xCC, 0x7C, 0x0C, 0x18, 0x70, 0x00 },  // 0x39 (nine)
    { 0x00, 0x30, 0x30, 0x00, 0x00, 0x30, 0x30, 0x00 },  // 0x3A (colon)
    { 0x00, 0x30, 0x30, 0x00, 0x30, 0x30, 0x60, 0x00 },  // 0x3B (semicolon)
    { 0x18, 0x30, 0x60, 0xC0, 0x60, 0x30, 0x18, 0x00 },  // 0x3C (less)
    { 0x00, 0x00, 0xFC, 0x00, 0x00, 0xFC, 0x00, 0x00 },  // 0x3D (equal)
    { 0x60, 0x30, 0x18, 0x0C, 0x18, 0x30, 0x60, 0x00 },  // 0x3E (greater)
    { 0x78, 0xCC, 0x0C, 0x18, 0x30, 0x00, 0x30, 0x00 },  // 0x3F (question)
    { 0x7C, 0xC6, 0xDE, 0xDE, 0xDE, 0xC0, 0x78, 0x00 },  // 0x40 (at)
    { 0x30, 0x78, 0xCC, 0xCC, 0xFC, 0xCC, 0xCC, 0x00 },  // 0x41 (A)
    { 0xFC, 0x66, 0x66, 0x7C, 0x66, 0x66, 0xFC, 0x00 },  // 0x42 (B)
    { 0x3C, 0x66, 0xC0, 0xC0, 0xC0, 0x66, 0x3C, 0x00 },  // 0x43 (C)
    { 0xF8, 0x6C, 0x66, 0x66, 0x66, 0x6C, 0xF8, 0x00 },  // 0x44 (D)
    { 0xFE, 0x62, 0x68, 0x78, 0x68, 0x62, 0xFE, 0x00 },  // 0x45 (E)
    { 0xFE, 0x62, 0x68, 0x78, 0x68, 0x60, 0xF0, 0x00 },  // 0x46 (F)
    { 0x3C, 0x66, 0xC0, 0xC0, 0xCE, 0x66, 0x3E, 0x00 },  // 0x47 (G)
    { 0xCC, 0xCC, 0xCC, 0xFC, 0xCC, 0xCC, 0xCC, 0x00 },  // 0x48 (H)
    { 0x78, 0x30, 0x30, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0x49 (I)
    { 0x1E, 0x0C, 0x0C, 0x0C, 0xCC, 0xCC, 0x78, 0x00 },  // 0x4A (J)
    { 0xE6, 0x66, 0x6C, 0x78, 0x6C, 0x66, 0xE6, 0x00 },  // 0x4B (K)
    { 0xF0, 0x60, 0x60, 0x60, 0x62, 0x66, 0xFE, 0x00 },  // 0x4C (L)
    { 0xC6, 0xEE, 0xFE, 0xFE, 0xD6, 0xC6, 0xC6, 0x00 },  // 0x4D (M)
    { 0xC6, 0xE6, 0xF6, 0xDE, 0xCE, 0xC6, 0xC6, 0x00 },  // 0x4E (N)
    { 0x38, 0x6C, 0xC6, 0xC6, 0xC6, 0x6C, 0x38, 0x00 },  // 0x4F (O)
    { 0xFC, 0x66, 0x66, 0x7C, 0x60, 0x60, 0xF0, 0x00 },  // 0x50 (P)
    { 0x78, 0xCC, 0xCC, 0xCC, 0xDC, 0x78, 0x1C, 0x00 },  // 0x51 (Q)
    { 0xFC, 0x66, 0x66, 0x7C, 0x6C, 0x66, 0xE6, 0x00 },  // 0x52 (R)
    { 0x78, 0xCC, 0xE0, 0x70, 0x1C, 0xCC, 0x78, 0x00 },  // 0x53 (S)
    { 0xFC, 0xB4, 0x30, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0x54 (T)
    { 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0xFC, 0x00 },  // 0x55 (U)
    { 0xCC, 0xCC, 0xCC, 0xCC, 0xCC, 0x78, 0x30, 0x00 },  // 0x56 (V)
    { 0xC6, 0xC6, 0xC6, 0xD6, 0xFE, 0xEE, 0xC6, 0x00 },  // 0x57 (W)
    { 0xC6, 0xC6, 0x6C, 0x38, 0x38, 0x6C, 0xC6, 0x00 },  // 0x58 (X)
    { 0xCC, 0xCC, 0xCC, 0x78, 0x30, 0x30, 0x78, 0x00 },  // 0x59 (Y)
    { 0xFE, 0xC6, 0x8C, 0x18, 0x32, 0x66, 0xFE, 0x00 },  // 0x5A (Z)
    { 0x78, 0x60, 0x60, 0x60, 0x60, 0x60, 0x78, 0x00 },  // 0x5B (bracketleft)
    { 0xC0, 0x60, 0x30, 0x18, 0x0C, 0x06, 0x02, 0x00 },  // 0x5C (backslash)
    { 0x78, 0x18, 0x18, 0x18, 0x18, 0x18, 0x78, 0x00 },  // 0x5D (bracketright)
    { 0x10, 0x38, 0x6C, 0xC6, 0x00, 0x00, 0x00, 0x00 },  // 0x5E (asciicircum)
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF },  // 0x5F (underscore)
    { 0x30, 0x30, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0x60 (grave)
    { 0x00, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x76, 0x00 },  // 0x61 (a)
    { 0xE0, 0x60, 0x60, 0x7C, 0x66, 0x66, 0xDC, 0x00 },  // 0x62 (b)
    { 0x00, 0x00, 0x78, 0xCC, 0xC0, 0xCC, 0x78, 0x00 },  // 0x63 (c)
    { 0x1C, 0x0C, 0x0C, 0x7C, 0xCC, 0xCC, 0x76, 0x00 },  // 0x64 (d)
    { 0x00, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },  // 0x65 (e)
    { 0x38, 0x6C, 0x60, 0xF0, 0x60, 0x60, 0xF0, 0x00 },  // 0x66 (f)
    { 0x00, 0x00, 0x76, 0xCC, 0xCC, 0x7C, 0x0C, 0xF8 },  // 0x67 (g)
    { 0xE0, 0x60, 0x6C, 0x76, 0x66, 0x66, 0xE6, 0x00 },  // 0x68 (h)
    { 0x30, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0x69 (i)
    { 0x0C, 0x00, 0x0C, 0x0C, 0x0C, 0xCC, 0xCC, 0x78 },  // 0x6A (j)
    { 0xE0, 0x60, 0x66, 0x6C, 0x78, 0x6C, 0xE6, 0x00 },  // 0x6B (k)
    { 0x70, 0x30, 0x30, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0x6C (l)
    { 0x00, 0x00, 0xCC, 0xFE, 0xFE, 0xD6, 0xC6, 0x00 },  // 0x6D (m)
    { 0x00, 0x00, 0xF8, 0xCC, 0xCC, 0xCC, 0xCC, 0x00 },  // 0x6E (n)
    { 0x00, 0x00, 0x78, 0xCC, 0xCC, 0xCC, 0x78, 0x00 },  // 0x6F (o)
    { 0x00, 0x00, 0xDC, 0x66, 0x66, 0x7C, 0x60, 0xF0 },  // 0x70 (p)
    { 0x00, 0x00, 0x76, 0xCC, 0xCC, 0x7C, 0x0C, 0x1E },  // 0x71 (q)
    { 0x00, 0x00, 0xDC, 0x76, 0x66, 0x60, 0xF0, 0x00 },  // 0x72 (r)
    { 0x00, 0x00, 0x7C, 0xC0, 0x78, 0x0C, 0xF8, 0x00 },  // 0x73 (s)
    { 0x10, 0x30, 0x7C, 0x30, 0x30, 0x34, 0x18, 0x00 },  // 0x74 (t)
    { 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0x76, 0x00 },  // 0x75 (u)
    { 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0x78, 0x30, 0x00 },  // 0x76 (v)
    { 0x00, 0x00, 0xC6, 0xD6, 0xFE, 0xFE, 0x6C, 0x00 },  // 0x77 (w)
    { 0x00, 0x00, 0xC6, 0x6C, 0x38, 0x6C, 0xC6, 0x00 },  // 0x78 (x)
    { 0x00, 0x00, 0xCC, 0xCC, 0xCC, 0x7C, 0x0C, 0xF8 },  // 0x79 (y)
    { 0x00, 0x00, 0xFC, 0x98, 0x30, 0x64, 0xFC, 0x00 },  // 0x7A (z)
    { 0x1C, 0x30, 0x30, 0xE0, 0x30, 0x30, 0x1C, 0x00 },  // 0x7B (braceleft)
    { 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x18, 0x00 },  // 0x7C (bar)
    { 0xE0, 0x30, 0x30, 0x1C, 0x30, 0x30, 0xE0, 0x00 },  // 0x7D (braceright)
    { 0x76, 0xDC, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0x7E (asciitilde)
    { 0x00, 0x10, 0x38, 0x6C, 0xC6, 0xC6, 0xFE, 0x00 },  // 0x7F (uni007F)
    // ---------------------------  160-255 ---------------------------
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0xA0 (uni00A0)
    { 0x18, 0x18, 0x00, 0x18, 0x18, 0x18, 0x18, 0x00 },  // 0xA1 (exclamdown)
    { 0x18, 0x18, 0x7E, 0xC0, 0xC0, 0x7E, 0x18, 0x18 },  // 0xA2 (cent)
    { 0x38, 0x6C, 0x64, 0xF0, 0x60, 0xE6, 0xFC, 0x00 },  // 0xA3 (sterling)
    { 0x00, 0xC6, 0x7C, 0xC6, 0xC6, 0x7C, 0xC6, 0x00 },  // 0xA4 (currency)
    { 0xCC, 0xCC, 0x78, 0xFC, 0x30, 0xFC, 0x30, 0x30 },  // 0xA5 (yen)
    { 0x18, 0x18, 0x18, 0x00, 0x18, 0x18, 0x18, 0x00 },  // 0xA6 (brokenbar)
    { 0x3E, 0x63, 0x38, 0x6C, 0x6C, 0x38, 0xCC, 0x78 },  // 0xA7 (section)
    { 0xC6, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0xA8 (dieresis)
    { 0x7E, 0x81, 0x9D, 0xA1, 0xA1, 0x9D, 0x81, 0x7E },  // 0xA9 (copyright)
    { 0x3C, 0x6C, 0x6C, 0x3E, 0x00, 0x7E, 0x00, 0x00 },  // 0xAA (ordfeminine)
    { 0x00, 0x33, 0x66, 0xCC, 0x66, 0x33, 0x00, 0x00 },  // 0xAB (guillemotleft)
    { 0x00, 0x00, 0x00, 0xFC, 0x0C, 0x0C, 0x00, 0x00 },  // 0xAC (logicalnot)
    { 0x00, 0x00, 0x00, 0x7E, 0x00, 0x00, 0x00, 0x00 },  // 0xAD (uni00AD)
    { 0x7E, 0x81, 0xB9, 0xA5, 0xB9, 0xA5, 0x81, 0x7E },  // 0xAE (registered)
    { 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0xAF (macron)
    { 0x38, 0x6C, 0x6C, 0x38, 0x00, 0x00, 0x00, 0x00 },  // 0xB0 (degree)
    { 0x30, 0x30, 0xFC, 0x30, 0x30, 0x00, 0xFC, 0x00 },  // 0xB1 (plusminus)
    { 0x70, 0x18, 0x30, 0x60, 0x78, 0x00, 0x00, 0x00 },  // 0xB2 (uni00B2)
    { 0x70, 0x18, 0x30, 0x18, 0x70, 0x00, 0x00, 0x00 },  // 0xB3 (uni00B3)
    { 0x0C, 0x18, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00 },  // 0xB4 (acute)
    { 0x00, 0x66, 0x66, 0x66, 0x66, 0x7C, 0x60, 0xC0 },  // 0xB5 (mu)
    { 0x7F, 0xDB, 0xDB, 0x7B, 0x1B, 0x1B, 0x1B, 0x00 },  // 0xB6 (paragraph)
    { 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00 },  // 0xB7 (periodcentered)
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x0C, 0x38 },  // 0xB8 (cedilla)
    { 0x30, 0x70, 0x30, 0x30, 0x78, 0x00, 0x00, 0x00 },  // 0xB9 (uni00B9)
    { 0x38, 0x6C, 0x6C, 0x38, 0x00, 0x7C, 0x00, 0x00 },  // 0xBA (ordmasculine)
    { 0x00, 0xCC, 0x66, 0x33, 0x66, 0xCC, 0x00, 0x00 },  // 0xBB (guillemotright)
    { 0xC3, 0xC6, 0xCC, 0xDB, 0x37, 0x6F, 0xCF, 0x03 },  // 0xBC (onequarter)
    { 0xC3, 0xC6, 0xCC, 0xDE, 0x33, 0x66, 0xCC, 0x0F },  // 0xBD (onehalf)
    { 0xE0, 0x33, 0x66, 0x3C, 0xFB, 0x37, 0x6F, 0xC3 },  // 0xBE (threequarters)
    { 0x30, 0x00, 0x30, 0x60, 0xC0, 0xCC, 0x78, 0x00 },  // 0xBF (questiondown)
    { 0xC0, 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0xC6, 0x00 },  // 0xC0 (Agrave)
    { 0x06, 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0xC6, 0x00 },  // 0xC1 (Aacute)
    { 0x7C, 0xC6, 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0x00 },  // 0xC2 (Acircumflex)
    { 0x76, 0xDC, 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0x00 },  // 0xC3 (Atilde)
    { 0xC6, 0x38, 0x6C, 0xC6, 0xFE, 0xC6, 0xC6, 0x00 },  // 0xC4 (Adieresis)
    { 0x30, 0x30, 0x00, 0x78, 0xCC, 0xFC, 0xCC, 0x00 },  // 0xC5 (Aring)
    { 0x3E, 0x6C, 0xCC, 0xFE, 0xCC, 0xCC, 0xCE, 0x00 },  // 0xC6 (AE)
    { 0x78, 0xCC, 0xC0, 0xCC, 0x78, 0x18, 0x0C, 0x78 },  // 0xC7 (Ccedilla)
    { 0xE0, 0x00, 0xFC, 0x60, 0x78, 0x60, 0xFC, 0x00 },  // 0xC8 (Egrave)
    { 0x1C, 0x00, 0xFC, 0x60, 0x78, 0x60, 0xFC, 0x00 },  // 0xC9 (Eacute)
    { 0x78, 0xCC, 0xFC, 0x60, 0x78, 0x60, 0xFC, 0x00 },  // 0xCA (Ecircumflex)
    { 0xCC, 0x00, 0xFC, 0x60, 0x78, 0x60, 0xFC, 0x00 },  // 0xCB (Edieresis)
    { 0xE0, 0x00, 0x78, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0xCC (Igrave)
    { 0x1C, 0x00, 0x78, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0xCD (Iacute)
    { 0x7E, 0xC3, 0x3C, 0x18, 0x18, 0x18, 0x3C, 0x00 },  // 0xCE (Icircumflex)
    { 0xCC, 0x00, 0x78, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0xCF (Idieresis)
    { 0xF8, 0x6C, 0x66, 0xF6, 0x66, 0x6C, 0xF8, 0x00 },  // 0xD0 (Eth)
    { 0xFC, 0x00, 0xCC, 0xEC, 0xFC, 0xDC, 0xCC, 0x00 },  // 0xD1 (Ntilde)
    { 0xC0, 0x38, 0x6C, 0xC6, 0xC6, 0x6C, 0x38, 0x00 },  // 0xD2 (Ograve)
    { 0x06, 0x38, 0x6C, 0xC6, 0xC6, 0x6C, 0x38, 0x00 },  // 0xD3 (Oacute)
    { 0x7C, 0xC6, 0x38, 0x6C, 0xC6, 0x6C, 0x38, 0x00 },  // 0xD4 (Ocircumflex)
    { 0x76, 0xDC, 0x38, 0x6C, 0xC6, 0x6C, 0x38, 0x00 },  // 0xD5 (Otilde)
    { 0xC3, 0x18, 0x3C, 0x66, 0x66, 0x3C, 0x18, 0x00 },  // 0xD6 (Odieresis)
    { 0x00, 0xC6, 0x6C, 0x38, 0x6C, 0xC6, 0x00, 0x00 },  // 0xD7 (multiply)
    { 0x3A, 0x6C, 0xCE, 0xD6, 0xE6, 0x6C, 0xB8, 0x00 },  // 0xD8 (Oslash)
    { 0xE0, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0x78, 0x00 },  // 0xD9 (Ugrave)
    { 0x1C, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0x78, 0x00 },  // 0xDA (Uacute)
    { 0x7C, 0xC6, 0x00, 0xC6, 0xC6, 0xC6, 0x7C, 0x00 },  // 0xDB (Ucircumflex)
    { 0xCC, 0x00, 0xCC, 0xCC, 0xCC, 0xCC, 0x78, 0x00 },  // 0xDC (Udieresis)
    { 0x1C, 0x00, 0xCC, 0xCC, 0x78, 0x30, 0x78, 0x00 },  // 0xDD (Yacute)
    { 0xF0, 0x60, 0x7C, 0x66, 0x7C, 0x60, 0xF0, 0x00 },  // 0xDE (Thorn)
    { 0x78, 0xCC, 0xCC, 0xD8, 0xCC, 0xC6, 0xCC, 0x00 },  // 0xDF (germandbls)
    { 0xE0, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },  // 0xE0 (agrave)
    { 0x1C, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },  // 0xE1 (aacute)
    { 0x7E, 0xC3, 0x3C, 0x06, 0x3E, 0x66, 0x3F, 0x00 },  // 0xE2 (acircumflex)
    { 0x76, 0xDC, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },  // 0xE3 (atilde)
    { 0xCC, 0x00, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },  // 0xE4 (adieresis)
    { 0x30, 0x30, 0x78, 0x0C, 0x7C, 0xCC, 0x7E, 0x00 },  // 0xE5 (aring)
    { 0x00, 0x00, 0x7F, 0x0C, 0x7F, 0xCC, 0x7F, 0x00 },  // 0xE6 (ae)
    { 0x00, 0x00, 0x78, 0xC0, 0xC0, 0x78, 0x0C, 0x38 },  // 0xE7 (ccedilla)
    { 0xE0, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },  // 0xE8 (egrave)
    { 0x1C, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },  // 0xE9 (eacute)
    { 0x7E, 0xC3, 0x3C, 0x66, 0x7E, 0x60, 0x3C, 0x00 },  // 0xEA (ecircumflex)
    { 0xCC, 0x00, 0x78, 0xCC, 0xFC, 0xC0, 0x78, 0x00 },  // 0xEB (edieresis)
    { 0xE0, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0xEC (igrave)
    { 0x38, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0xED (iacute)
    { 0x7C, 0xC6, 0x38, 0x18, 0x18, 0x18, 0x3C, 0x00 },  // 0xEE (icircumflex)
    { 0xCC, 0x00, 0x70, 0x30, 0x30, 0x30, 0x78, 0x00 },  // 0xEF (idieresis)
    { 0x30, 0x7E, 0x0C, 0x7C, 0xCC, 0xCC, 0x78, 0x00 },  // 0xF0 (eth)
    { 0x00, 0xF8, 0x00, 0xF8, 0xCC, 0xCC, 0xCC, 0x00 },  // 0xF1 (ntilde)
    { 0x00, 0xE0, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },  // 0xF2 (ograve)
    { 0x00, 0x1C, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },  // 0xF3 (oacute)
    { 0x78, 0xCC, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },  // 0xF4 (ocircumflex)
    { 0x76, 0xDC, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },  // 0xF5 (otilde)
    { 0x00, 0xCC, 0x00, 0x78, 0xCC, 0xCC, 0x78, 0x00 },  // 0xF6 (odieresis)
    { 0x30, 0x30, 0x00, 0xFC, 0x00, 0x30, 0x30, 0x00 },  // 0xF7 (divide)
    { 0x00, 0x02, 0x7C, 0xCE, 0xD6, 0xE6, 0x7C, 0x80 },  // 0xF8 (oslash)
    { 0x00, 0xE0, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },  // 0xF9 (ugrave)
    { 0x00, 0x1C, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },  // 0xFA (uacute)
    { 0x78, 0xCC, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },  // 0xFB (ucircumflex)
    { 0x00, 0xCC, 0x00, 0xCC, 0xCC, 0xCC, 0x7E, 0x00 },  // 0xFC (udieresis)
    { 0x00, 0x1C, 0x00, 0xCC, 0xCC, 0x7C, 0x0C, 0x78 },  // 0xFD (yacute)
    { 0xE0, 0x60, 0x7C, 0x66, 0x66, 0x7C, 0x60, 0xF0 },  // 0xFE (thorn)
    { 0x00, 0xCC, 0x00, 0xCC, 0xCC, 0x7C, 0x0C, 0xF8 },  // 0xFF (ydieresis)
};

void c8800retro_cfg_setup ( c8800retro_cfg_t *cfg ) 
{
    // Communication gpio pins
    cfg->scl  = HAL_PIN_NC;
    cfg->sda  = HAL_PIN_NC;

    // Additional gpio pins
    cfg->int_pin = HAL_PIN_NC;

    cfg->i2c_speed   = I2C_MASTER_SPEED_STANDARD;
    cfg->i2c_address = C8800RETRO_SET_DEV_ADDR;
}

err_t c8800retro_init ( c8800retro_t *ctx, c8800retro_cfg_t *cfg ) 
{
    i2c_master_config_t i2c_cfg;

    i2c_master_configure_default( &i2c_cfg );

    i2c_cfg.scl = cfg->scl;
    i2c_cfg.sda = cfg->sda;

    ctx->slave_address = cfg->i2c_address;

    if ( i2c_master_open( &ctx->i2c, &i2c_cfg ) == I2C_MASTER_ERROR ) 
    {
        return I2C_MASTER_ERROR;
    }

    if ( i2c_master_set_slave_address( &ctx->i2c, ctx->slave_address ) == I2C_MASTER_ERROR ) 
    {
        return I2C_MASTER_ERROR;
    }

    if ( i2c_master_set_speed( &ctx->i2c, cfg->i2c_speed ) == I2C_MASTER_ERROR ) 
    {
        return I2C_MASTER_ERROR;
    }

    digital_in_init( &ctx->int_pin, cfg->int_pin );

    return I2C_MASTER_SUCCESS;
}

err_t c8800retro_default_cfg ( c8800retro_t *ctx ) 
{
    err_t error_flag = c8800retro_write_reg( ctx, C8800RETRO_REG_DECODE_MODE, C8800RETRO_NO_DECODE );
    error_flag |= c8800retro_write_reg( ctx, C8800RETRO_REG_GLOBAL_INTENSITY, C8800RETRO_INTENSITY_8 );
    error_flag |= c8800retro_write_reg( ctx, C8800RETRO_REG_SCAN_LIMIT, C8800RETRO_SCAN_LIMIT_0_7 );
    error_flag |= c8800retro_write_reg( ctx, C8800RETRO_REG_SHUTDOWN, C8800RETRO_NORMAL_OPERATION_RESET );
    Delay_100ms( );
    error_flag |= c8800retro_clear_display( ctx );
    error_flag |= c8800retro_clear_interrupt( ctx );
    
    return error_flag;
}

err_t c8800retro_write_reg ( c8800retro_t *ctx, uint8_t reg, uint8_t data_in )
{
    uint8_t tx_buf[ 2 ];

    if ( reg < C8800RETRO_REG_DIGIT_0 || 
       ( reg > C8800RETRO_REG_KEYB && reg != C8800RETRO_REG_SELF_ADDRESSING ) )
    {
        return C8800RETRO_ERROR;
    }
    
    tx_buf[ 0 ] = reg;
    tx_buf[ 1 ] = data_in;
    
    return i2c_master_write( &ctx->i2c, tx_buf, 2 );
}

err_t c8800retro_read_reg ( c8800retro_t *ctx, uint8_t reg, uint8_t *data_out, uint8_t len )
{
    if ( reg < C8800RETRO_REG_DIGIT_0 || 
         ( reg > C8800RETRO_REG_KEYB && reg != C8800RETRO_REG_SELF_ADDRESSING ) )
    {
        return C8800RETRO_ERROR;
    }

    return i2c_master_write_then_read( &ctx->i2c, &reg, 1, data_out, len );;
}

err_t c8800retro_clear_display ( c8800retro_t *ctx )
{  
    err_t error_flag = C8800RETRO_OK;
    
    for ( uint8_t cnt = C8800RETRO_REG_DIGIT_0; cnt <= C8800RETRO_REG_DIGIT_7; cnt++ )
    {
        error_flag |= c8800retro_write_reg( ctx, cnt, 0x00 );
    }
    return error_flag;
}

err_t c8800retro_display_char ( c8800retro_t *ctx, uint8_t char_index )
{
    uint8_t start_reg = C8800RETRO_REG_DIGIT_0;
    err_t error_flag = C8800RETRO_OK;
    
    if ( char_index > 127 )
    {
        if ( char_index >= 160 )
        {
            char_index -= 32;
        }
        else
        {
            return C8800RETRO_ERROR;
        }
    }

    for ( uint8_t cnt = 0; cnt < 8; cnt++ )
    {
        error_flag |= c8800retro_write_reg( ctx, start_reg++, font8x8_ib8x8u[ char_index ][ cnt ] );
    }
    return error_flag;
}

err_t c8800retro_display_image ( c8800retro_t *ctx, uint8_t *p_image )
{
    uint8_t start_reg = C8800RETRO_REG_DIGIT_0;
    err_t error_flag = C8800RETRO_OK;
    
    if ( p_image == NULL )
    {
        return C8800RETRO_ERROR;
    }

    for ( uint8_t cnt = 0; cnt < 8; cnt++ )
    {
        error_flag |= c8800retro_write_reg( ctx, start_reg++, p_image[ cnt ] );
    }
    return error_flag;
}

err_t c8800retro_display_string ( c8800retro_t *ctx, uint8_t scroll_mode, uint16_t scroll_speed_ms, uint8_t *text )
{    
    uint8_t char_buf[ 256 ];
    uint8_t tmp_char_buf[ 8 ];
    uint8_t tmp_char = 0;
    uint16_t buf_len = 0;
    uint16_t buf_cnt = 0;
    uint16_t cnt = 0;
    err_t error_flag = C8800RETRO_OK;
    
    if ( text == NULL )
    {
        return C8800RETRO_ERROR;
    }
    
    for ( buf_cnt = 0; buf_cnt < strlen( text ); buf_cnt++ )
    {
        tmp_char = text[ buf_cnt ];
        for ( cnt = 0; cnt < 8; cnt++ )
        {
            char_buf[ buf_len++ ] = font8x8_ib8x8u[ tmp_char ][ cnt ];
        }
    }
    
    if ( C8800RETRO_SCROLL_HORIZONTAL_LEFT == scroll_mode || 
         C8800RETRO_SCROLL_VERTICAL_UP == scroll_mode )
    {
        for ( buf_cnt = 0; buf_cnt < ( buf_len - 8 ); buf_cnt++ )
        {
            if ( C8800RETRO_SCROLL_HORIZONTAL_LEFT == scroll_mode )
            {
                for ( cnt = 0; cnt < 8; cnt++ )
                {
                    tmp_char_buf[ cnt ] = ( char_buf[ buf_cnt - buf_cnt % 8 + cnt ] << ( buf_cnt % 8 ) ) | 
                                        ( char_buf[ buf_cnt - buf_cnt % 8 + cnt + 8 ] >> ( 8 - buf_cnt % 8 ) );
                }
                error_flag |= c8800retro_display_image( ctx, tmp_char_buf );
            }
            else if ( C8800RETRO_SCROLL_VERTICAL_UP == scroll_mode )
            {
                error_flag |= c8800retro_display_image( ctx, &char_buf[ buf_cnt ] );
            }
            
            for ( cnt = 0; cnt < scroll_speed_ms; cnt++ )
            {
                Delay_1ms( );
            }
        }
    }
    
    if ( C8800RETRO_SCROLL_HORIZONTAL_RIGHT == scroll_mode || 
         C8800RETRO_SCROLL_VERTICAL_DOWN == scroll_mode )
    {
        for ( buf_cnt = buf_len; buf_cnt >= 8; buf_cnt-- )
        {
            if ( C8800RETRO_SCROLL_HORIZONTAL_RIGHT == scroll_mode )
            {
                for ( cnt = 0; cnt < 8; cnt++ )
                {
                    tmp_char_buf[ cnt ] = ( char_buf[ buf_cnt - buf_cnt % 8 + cnt - 8 ] << ( buf_cnt % 8 ) ) | 
                                        ( char_buf[ buf_cnt - buf_cnt % 8 + cnt ] >> ( 8 - buf_cnt % 8 ) );
                }
                error_flag |= c8800retro_display_image( ctx, tmp_char_buf );
            }
            else if ( C8800RETRO_SCROLL_VERTICAL_DOWN == scroll_mode )
            {
                error_flag |= c8800retro_display_image( ctx, &char_buf[ buf_cnt - 8 ] );
            }
            
            for ( cnt = 0; cnt < scroll_speed_ms; cnt++ )
            {
                Delay_1ms( );
            }
        }
    }
    
    
    return error_flag;
}

uint8_t c8800retro_get_int_pin ( c8800retro_t *ctx )
{
    return digital_in_read( &ctx->int_pin );
}

err_t c8800retro_clear_interrupt ( c8800retro_t *ctx )
{  
    uint8_t tmp_data[ 2 ] = { 0 }; 
    
    return c8800retro_read_reg( ctx, C8800RETRO_REG_KEYA, tmp_data, 2 );
}

err_t c8800retro_get_keypad ( c8800retro_t *ctx, uint16_t *keypad )
{  
    uint8_t tmp_data[ 2 ] = { 0 }; 
    
    if ( C8800RETRO_ERROR == c8800retro_read_reg( ctx, C8800RETRO_REG_KEYA, tmp_data, 2 ) )
    {
        return C8800RETRO_ERROR;
    }
    
    *keypad = ( ( uint16_t ) tmp_data[ 0 ] << 8 ) | tmp_data[ 1 ];
    *keypad = ~*keypad;
    
    return C8800RETRO_OK;
}

// ------------------------------------------------------------------------- END
